# EC2 user data script to run nshmp-haz.
#
# Global variables come from nshmp-haz-s3-lambda as the Lambda function creates the EC2 instance.

set -o errexit;
set -o errtrace;

# EC2 home path
readonly HOME="/home/ec2-user";

# Instance id, set in ec2_setup
INSTANCE_ID="";

# Log file
readonly LOG_FILE="ec2-user-data.log";

readonly S3_OUTPUT_URL="s3://${S3_BUCKET}/${S3_OUTPUT_PATH}/${S3_OUTPUT_DIR}";

####
# Run nshmp-haz code.
# Globals:
#   (string) HOME - EC2 home path
#   (boolean) ACCESS_VISUALVM - Whether to gain access to Java Visual VM
#   (string) AWS_REGION - The AWS region
#   (string) NSHMP_SITE_FILE_NAME - Site file name to run nshmp-haz with
#   (string) NSHMP_CONFIG_FILE_NAME - The config file name to run nshmp-haz with
#   (string) NSHMP_MODEL - The NSHM
#   (string) NSHMP_PROGRAM - The program to run: deagg | deagg-epsilon | hazard | rate
#   (number) RETURN_PERIOD - The return period for deagg
#   (number) IML - The intensity measure level for deagg-iml
#   (string) JAVA_XMX - Max memory allocation for Java
#   (boolean) SEND_UPDATES - Whether to send periodic updates
#   (string) S3_OUTPUT_DIR - The S3 bucket directory to put output
#   (string) INSTANCE_ID - The EC2 instance id
#   (string) LOG_FILE - The log file
# Arguments:
#   None
# Returns:
#   None
####
main() {
  # Set trap for uncaught errors.
  trap 'error_exit "${BASH_COMMAND}" "$(< ${LOG_FILE})"' ERR;

  # Setup EC2
  ec2_setup 2> ${LOG_FILE};

  # Change directory home
  cd ${HOME} 2> ${LOG_FILE};

  # Set region
  aws configure set region ${AWS_REGION} 2> ${LOG_FILE};

  # Start awslogs
  start_awslogs 2> ${LOG_FILE};

  # Get Docker images
  get_docker_images 2> ${LOG_FILE};

  # Write site file
  get_site_file 2> ${LOG_FILE};

  # Write config file
  get_config_file 2> ${LOG_FILE};

  # Monitor log file
  tail -f ${LOG_FILE} &

  # Set watch job
  set_watch_job 2> ${LOG_FILE};

  # Set Java VisualVM opts
  local visualvm_opts=""
  if [ ${ACCESS_VISUALVM} = true ]; then
    visualvm_opts="
      -p 9090:9090
      -e ACCESS_VISUALVM=${ACCESS_VISUALVM}
    ";
  fi

  # Message to send at start up
  local message="
    Starting nshmp-haz ${NSHMP_PROGRAM}.

    ----------
    Trigger config file:

    \n${TRIGGER_CONFIG_FILE}

  ";

  # Send message that nshmp-haz has started
  send_notification "${message}" 2> ${LOG_FILE};

  # Download custom model
  if [ ${USE_CUSTOM_MODEL} = true ]; then
    mkdir model;
    cd model;
    curl -L ${NSHMP_MODEL} | tar -x --strip-components=1;
    cd ..;
  fi

  # Run nshmp-haz
  docker run \
      -e PROGRAM=${NSHMP_PROGRAM} \
      -e MODEL=${NSHMP_MODEL} \
      -e JAVA_XMS=${JAVA_XMX} \
      -e JAVA_XMX=${JAVA_XMX} \
      -e RETURN_PERIOD=${RETURN_PERIOD}  \
      -e IML=${IML} \
      -e MOUNT_MODEL=${USE_CUSTOM_MODEL} \
      -v $(pwd)/model:/app/model \
      -v $(pwd)/${NSHMP_SITE_FILE_NAME}:/app/sites.${NSHMP_SITE_FILE_NAME##*.} \
      -v $(pwd)/${NSHMP_CONFIG_FILE_NAME}:/app/config.json \
      -v $(pwd)/${S3_OUTPUT_DIR}:/app/output \
      ${visualvm_opts} \
      --network="host" \
      usgs/nshmp-haz 1> ${LOG_FILE} ||
      error_exit \
          "Failed to run nshmp-haz" \
          "$(tail -55 ${LOG_FILE})";

  # Kill watch job
  kill_watch_job 2> ${LOG_FILE};

  # Create lambda config file
  echo "${TRIGGER_CONFIG_FILE}" > ${S3_OUTPUT_DIR}/${TRIGGER_CONFIG_FILE_NAME} 2> ${LOG_FILE};

  # Move output to S3
  mv_output_to_s3 2> ${LOG_FILE};

  # Create slices
  create_slices 2> ${LOG_FILE};

  # Update S3 bucket
  update_s3 2> ${LOG_FILE};

  # Update results metadata
  aws lambda invoke \
      --cli-read-timeout 600 \
      --function-name 'nshmp-haz-results-metadata' haz-results.json 2> ${LOG_FILE} || \
      send_warning "AWS Lambda [nshmp-haz-results-metadata]" "$(< ${LOG_FILE})";

  # Run complete
  exit_complete;
}

####
# Run nshmp-hazard-results-slicer Lambda function.
# Globals:
#   (string) S3_BUCKET - The S3 bucket for output
#   (string) S3_OUTPUT_DIR - The S3 bucket directory to put output
#   (string) LOG_FILE - The log file
#   (number[]) SLICES - The slices for hazard map
# Arguments:
#   None
# Returns:
#   None
####
create_slices() {
  local query="query.json";
  printf "
  {
    \"bucket\": \"${S3_BUCKET}\",
    \"key\": \"${S3_OUTPUT_PATH}/${S3_OUTPUT_DIR}\",
    \"slices\": [${SLICES}]
  }
  " > ${query} 2> ${LOG_FILE};

  local lambda_output="lambda-output.json";

  aws lambda invoke \
      --function-name nshmp-haz-results-slicer \
      --payload file://${query} ${lambda_output} 2> ${LOG_FILE} || \
      error_exit \
        "AWS Lambda [nshmp-haz-results-slicer] failed" \
        "$(< ${LOG_FILE})";

  if [ $(jq -r '.status' ${lambda_output}) != "success" ]; then
    local message=$(jq -r '.message' ${lambda_output}) 2> ${LOG_FILE};
    local request=$(jq -r '.request' ${lambda_output}) 2> ${LOG_FILE};
    error_exit \
        "AWS Lambda [nshmp-hazard-results-slicer] failed on S3 [${request}]" \
        "${message}";
  fi
}

####
# Install dependencies.
# Globals:
#   (string) HOME - EC2 home path
#   (string) JOB_ID - The job id
#   (string) INSTANCE_ID - EC2 instance id
#   (string) LOG_FILE - The log file
#   (string) SNS_TOPIC_ARN - SNS topic arn
#   (string) TRIGGER_CONFIG_FILE - The trigger file
#   (function) error_exit - Exit on error function
#   (function) send_notification - Send notification function
#   (function) shutdown_ec2 - Shutdown EC2 function
# Arguments:
#   None
# Returns:
#   None
####
ec2_setup() {
  yum update -y;
  yum install -y awslogs curl jq zip;

  # Set instance id
  INSTANCE_ID=$(curl http://169.254.169.254/latest/meta-data/instance-id);

  # Set LANG for logs
  export LANG=en_US.UTF-8;

  # Set terminal, needed for watch command
  export TERM=xterm-256color;

  # Exports for ec2-send-updates
  export HOME;
  export JOB_ID;
  export INSTANCE_ID;
  export LOG_FILE;
  export SNS_TOPIC_ARN;
  export TRIGGER_CONFIG_FILE;
  export -f error_exit;
  export -f send_notification;
  export -f shutdown_ec2;
}

####
# Shutdown EC2 and exit code status -1
# Globals:
#   (string) TRIGGER_CONFIG_FILE - The trigger config file
#   (string) INSTANCE_ID - The EC2 instance id
# Arguments:
#   (string) message - Message to send via send_notification
#   (string) log - The log to send
# Returns:
#   None
####
error_exit() {
  # Kill watch job
  kill_watch_job 2> ${LOG_FILE};

  local message="
    AWS EC2 nshmp error (instance id - ${INSTANCE_ID}):
    ${1}

    ----------
    Logs:

    ${2}

    ----------
    Trigger config file:

    \n${TRIGGER_CONFIG_FILE}

  ";

  send_notification "${message}" || shutdown_ec2;
  shutdown_ec2;
  exit -1;
}

####
# nshmp-haz run complete. Send notification and shutdown.
# Globals:
#   (string) AWS_REGION - The AWS region
#   (string) S3_BUCKET - The S3 bucket for output
#   (string) S3_OUTPUT_DIR - The S3 bucket directory to put output
#   (string) INSTANCE_ID - The EC2 instance id
# Arguments:
#   None
# Returns:
#   None
####
exit_complete() {
  local bucket_url="https://s3-${AWS_REGION}.amazonaws.com/${S3_BUCKET}/${S3_OUTPUT_PATH}";

  local message="
    nshmp-haz complete (instance id - ${INSTANCE_ID})

    Results are in S3 bucket ${S3_BUCKET}/${S3_OUTPUT_PATH}/${S3_OUTPUT_DIR}

    Download: ${bucket_url}/${S3_OUTPUT_DIR}/${S3_OUTPUT_DIR}.zip
  ";

  send_notification "${message}" || shutdown_ec2;
  shutdown_ec2;
  exit 0;
}

####
# Get the config file.
# Globals:
#   (string) NSHMP_CONFIG_FILE - The URL to the config file
#   (string) NSHMP_CONFIG_FILE_NAME - The config file name and extension
#   (string) LOG_FILE - The log file
# Arguments:
#   None
# Returns:
#   None
####
get_config_file() {
  if [ ${NSHMP_CONFIG_FILE} == null ]; then
    echo "{}" > ${NSHMP_CONFIG_FILE_NAME};
  else
    curl -L -o ${NSHMP_CONFIG_FILE_NAME} ${NSHMP_CONFIG_FILE};
  fi

  # Check if file is valid JSON
  jq empty < ${NSHMP_CONFIG_FILE_NAME} 2> ${LOG_FILE} || \
      error_exit "Config file [${NSHMP_CONFIG_FILE_NAME}] is not valid JSON" "$(< ${LOG_FILE})";
}

####
# Get nshmp-haz and nshmp-site-ws Docker images. Start nshmp-site-ws on port 8080.
# Globals:
#   (string) LOG_FILE - The log file
# Arguments:
#   None
# Returns:
#   None
####
get_docker_images() {
  # Get usgs/nshmp-haz Docker image
  docker pull usgs/nshmp-haz 2> ${LOG_FILE} || \
      error_exit "Could not get nshmp-haz Docker image" "$(< ${LOG_FILE})";

  # Get usgsnshmp/nshmp-site-ws image
  docker pull usgsnshmp/nshmp-site-ws 2> ${LOG_FILE} || \
      error_exit "Could not get nshmp-site-ws Docker image" "$(< ${LOG_FILE})";

  # Run basin service
  docker run -p 8080:8080 -d usgsnshmp/nshmp-site-ws 2> ${LOG_FILE} || \
      error_exit "Could not start nshmp-site-ws" "$(< ${LOG_FILE})";
}

####
# Get the site file.
# Globals:
#   (string) NSHMP_SITE_FILE - The URL to the site file
#   (string) NSHMP_SITE_FILE_NAME - The site file name and extension
#   (string) LOG_FILE - The log file
# Arguments:
#   None
# Returns:
#   None
####
get_site_file() {
  # Download sites file
  curl -L -o ${NSHMP_SITE_FILE_NAME} ${NSHMP_SITE_FILE};

  # Check if valid JSON or ASCII file
  case ${NSHMP_SITE_FILE_NAME} in
    *geojson)
      jq empty < ${NSHMP_SITE_FILE_NAME} 2> ${LOG_FILE} || \
          error_exit "Site file [${NSHMP_SITE_FILE_NAME}] is not valid JSON" "$(< ${LOG_FILE})";
      ;;
    *csv)
      if [[ "$(file ${NSHMP_SITE_FILE_NAME} -b)" != "ASCII text"* ]]; then
        error_exit \
            "Site file [${NSHMP_SITE_FILE_NAME}] is not valud ASCII" \
            "Site file is not valid ASCII";
      fi
      ;;
    *)
      error_exit \
        "Site file [${NSHMP_SITE_FILE_NAME}] extension not supported." \
        "Bad site file extension. Only csv or geojson supported.";
      ;;
  esac
}

####
# Kill watch job.
# Globals:
#   (string) LOG_FILE - The log file
#   (boolean) SEND_UPDATES - Whether to send periodic updates
# Arguments:
#   None
# Return:
#   None
####
kill_watch_job() {
  # Kill watch job
  if [ ${SEND_UPDATES} == true ]; then
    kill -INT $(pidof watch) 2> ${LOG_FILE} || printf "\nWARNING: Could not stop watch job \n";
  fi
}

####
# Move output directory to S3.
# Globals:
#   (string) S3_BUCKET - The S3 bucket for output
#   (string) S3_OUTPUT_DIR - The S3 bucket directory to put output
#   (string) LOG_FILE - The log file
# Arguments:
#   None
# Returns:
#   None
####
mv_output_to_s3() {
  aws s3 cp --acl public-read \
      ${S3_OUTPUT_DIR} ${S3_OUTPUT_URL} --recursive 2> ${LOG_FILE} || \
      error_exit \
          "Failed copying to S3 bucket [${S3_BUCKET}/output/${S3_OUTPUT_DIR}]" \
          "$(< ${LOG_FILE})";
}

####
# Send notification via email to a specific SNS topic.
# Globals:
#   (string) HOME - EC2 home path
#   (string) SNS_TOPIC_ARN - The SNS topic ARN
#   (string) JOB_ID - The job ID
# Arguments:
#   (string) message - The message to send
# Returns:
#   None
####
send_notification() {
  printf "${1}" > ${HOME}/email.txt || shutdown_ec2;

  aws sns publish \
      --topic-arn ${SNS_TOPIC_ARN} \
      --subject "AWS nshmp-haz (job id: ${JOB_ID})" \
      --message file://${HOME}/email.txt || shutdown_ec2;
}

####
# Send a warning.
# Globals:
#   (string) INSTANCE_ID - The EC2 instance id
# Arguments:
#   (string) message - Message to send via send_notification
#   (string) log - The log to send
# Returns:
#   None
####
send_warning() {
  local message="
    AWS EC2 nshmp warning (instance id - ${INSTANCE_ID}):
    ${1}

    ----------
    Logs:

    ${2}
  ";

  send_notification "${message}" || shutdown_ec2;
}

####
# Shutdown the EC2 instance.
# Globals:
#   None
# Arguments:
#   None
# Returns:
#   None
####
shutdown_ec2() {
  shutdown -h +1;
}

####
# Start up awslogs, stream logs to CloudWatch.
# Globals:
#   (string) AWS_REGION - The AWS region
#   (string) JOB_ID - The job id
#   (string) INSTANCE_ID - The instance id
#   (string) EMAIL - The email of user running nshmp-haz
#   (string) TRIGGER_CONFIG_FILE_NAME - The trigger file name
# Arguments:
#   None
# Returns:
#   None
####
start_awslogs() {
  local ec2_name="nshmp-haz-${JOB_ID}";
  local stream_name="${ec2_name}/${INSTANCE_ID}_${EMAIL}_${TRIGGER_CONFIG_FILE_NAME}";

  echo "region = ${AWS_REGION}" >> /etc/awslogs/awscli.conf;
  echo "log_group_name = nshmp-haz-ec2-logs" >> /etc/awslogs/awslogs.conf;
  echo "log_stream_name = ${stream_name}" >> /etc/awslogs/awslogs.conf;
  systemctl start awslogsd;
}

####
# Get the Bash script to email updates and schedule it with watch.
# Globals:
#   (string) HOME - EC2 home path
#   (string) SEND_UPDATES - Whether to send periodic updates
#   (string) SEND_UPDATES_EVERY - The time to send updates
#   (string) SNS_TOPIC_ARN - The SNS topic ARN
#   (string) JOB_ID - The job ID
# Arguments:
#   None
# Returns:
#   None
####
set_watch_job() {
  if [ ${SEND_UPDATES} == false ]; then
    return;
  fi

  local project="nshmp-haz-s3-lambda-trigger";
  local bash_path="src/lambdas/nshmp-haz/ec2-send-updates.sh";
  local git_lab_url="https://code.chs.usgs.gov/bclayton/${project}/";
  local archive_path="-/archive/master/${project}-master.tar.gz";
  local tar_url="${git_lab_url}${archive_path}";

  # Get tarball of project
  curl -L ${tar_url} | tar -xz;

  # Update project name
  mv ${project}-* ${project};

  # Schedule ec2-send-updates
  (sleep ${SEND_UPDATES_EVERY}; watch -n ${SEND_UPDATES_EVERY} \
      "bash ${HOME}/${project}/${bash_path} | tee -a ${HOME}/watch-log" &>/dev/null &) &
}


####
# Zip output directory and move to S3
# Globals:
#   (string) S3_OUTPUT_DIR - The S3 bucket directory to put output
#   (string) LOG_FILE - The log file
# Arguments:
#   None
# Returns:
#   None
####
update_s3() {
  rm -r ${S3_OUTPUT_DIR};
  aws s3 cp ${S3_OUTPUT_URL} ${S3_OUTPUT_DIR} --recursive 2> ${LOG_FILE};
  zip -r ${S3_OUTPUT_DIR}.zip ${S3_OUTPUT_DIR} 2> ${LOG_FILE};
  mv ${S3_OUTPUT_DIR}.zip ${S3_OUTPUT_DIR}/. 2> ${LOG_FILE};
  mv_output_to_s3 2> ${LOG_FILE};
}

####
# Run main.
####
main "$@" 2> ${LOG_FILE};
