import * as AWS from 'aws-sdk';
import { PromiseResult } from 'aws-sdk/lib/request';
import * as dateFormat from 'dateformat';

import { NshmpLambdaUtil, LambdaError } from './nshmp-lambda-util';
import { LambdaConfig } from '../config/lambda-config';

/**
 * Fire up an EC2 instace.
 */
export class NshmpHazS3Lambda {

  private ec2 = new AWS.EC2();
  private s3 = new AWS.S3();

  /**
   * Fire up an EC2 instance and run nshmp-haz
   *
   * @param config The config
   */
  async startComputeJob(config: LambdaConfig):
      Promise<PromiseResult<AWS.EC2.Reservation, AWS.AWSError> | void> {

    const nshmpConfig = config.triggerConfig.nshmp;
    const nshmpSiteFileName = nshmpConfig.nshmpSiteFile.split('/').pop();
    const nshmpConfigFileName = nshmpConfig.nshmpConfigFile != undefined ?
        nshmpConfig.nshmpConfigFile.split('/').pop() : 'config.json';

    const userDataScript = `
      #!/bin/bash

      # User data variables
      readonly ACCESS_VISUALVM=${config.triggerConfig.ec2.accessVisualVM};
      readonly SEND_UPDATES="${nshmpConfig.sendUpdates}";
      readonly SEND_UPDATES_EVERY="${nshmpConfig.sendUpdatesEvery}";
      readonly EMAIL="${nshmpConfig.email}";
      readonly NSHMP_CONFIG_FILE_NAME="${nshmpConfigFileName}";
      readonly NSHMP_CONFIG_FILE="${nshmpConfig.nshmpConfigFile || null}";
      readonly NSHMP_MODEL="${nshmpConfig.nshmpModel}";
      readonly NSHMP_SITE_FILE_NAME="${nshmpSiteFileName}";
      readonly NSHMP_SITE_FILE="${nshmpConfig.nshmpSiteFile}";
      readonly NSHMP_PROGRAM="${nshmpConfig.program}";
      readonly RETURN_PERIOD=${nshmpConfig.program == 'deagg' ? nshmpConfig.returnPeriod : ''};
      readonly IML=${nshmpConfig.program == 'deagg-iml' ? nshmpConfig.iml : ''};
      readonly JAVA_XMX="${config.triggerConfig.ec2.javaXmx}";
      readonly S3_BUCKET="${nshmpConfig.s3Bucket}";
      readonly S3_OUTPUT_PATH="${nshmpConfig.s3OutputPath}";
      readonly S3_OUTPUT_DIR="${nshmpConfig.s3OutputDirectory}-${config.jobId}";
      readonly SNS_TOPIC_ARN="${config.nshmp.snsTopic.TopicArn}";
      readonly AWS_REGION="${AWS.config.region}";
      readonly JOB_ID="${config.jobId}";
      readonly USE_CUSTOM_MODEL="${nshmpConfig.useCustomModel}";
      readonly SLICES=${nshmpConfig.slices};
      readonly TRIGGER_CONFIG_FILE_NAME="${config.triggerFileName}";
      readonly TRIGGER_CONFIG_FILE="${config.rawTriggerConfig}";

      # Run nshmp-haz
      ${config.ec2.ec2UserDataContents}
    `;

    return await this.checkInstance(config)
        .then(() => this.runEC2Instance(userDataScript, config));
  }

  private async checkInstance(config: LambdaConfig): Promise<void> {
      const results = await this.ec2.describeInstances({
        Filters: [
          {
            Name: 'tag:JobId',
            Values: [config.jobId]
          }
        ]
      }).promise();

      console.log('\nReservations: ', JSON.stringify(results, null, 2));
      if (results.Reservations.length > 0) {
        const reservation =  results.Reservations[0];
        if (reservation.Instances.length > 0) {
          throw new Error(`Duplicate jobId found [${config.jobId}]`);
        }
      }

      return;
    }


  private async runEC2Instance(userDataScript: string, config: LambdaConfig):
      Promise<PromiseResult<AWS.EC2.Reservation, AWS.AWSError> | void> {

    const nshmpConfig = config.triggerConfig.nshmp;
    const user = nshmpConfig.email.split('@').shift();
    const outputPath = `${nshmpConfig.s3Bucket}/${nshmpConfig.s3OutputPath}/`+
        `${nshmpConfig.s3OutputDirectory}-${config.jobId}`;

    return await this.ec2.runInstances({
      IamInstanceProfile: {
        Arn: config.ec2.ec2InstanceProfile
      },
      ImageId: config.ec2.ec2ImageId,
      InstanceInitiatedShutdownBehavior: 'terminate',
      InstanceType: config.triggerConfig.ec2.instanceType,
      KeyName: 'bclayton-dev-nshmp',
      MaxCount: 1,
      MinCount: 1,
      Monitoring: {
        Enabled: true
      },
      SecurityGroupIds: [
        'sg-0c807101c9a361169'
      ],
      SubnetId: 'subnet-08f1118dd59133513',
      TagSpecifications: [
        {
          ResourceType: 'instance',
          Tags: [
            {Key: 'User', Value: user},
            {Key: 'Email', Value: nshmpConfig.email},
            {Key: 'JobId', Value: config.jobId},
            {Key: 'Name', Value: `nshmp-haz-${config.jobId}`},
            {Key: 'nshmp:config-file', Value: nshmpConfig.nshmpConfigFile || ""},
            {Key: 'nshmp:model', Value: nshmpConfig.nshmpModel},
            {Key: 'nshmp:s3-output-path', Value: outputPath},
            {Key: 'nshmp:site-file', Value: nshmpConfig.nshmpSiteFile},
            {Key: 'Program', Value: `nshmp-haz ${nshmpConfig.program}`},
            {Key: 'TriggerFile', Value: config.triggerFileName},
            {Key: 'Purpose', Value: 'run nshmp-haz'},
            {Key: 'Source', Value: 'lambda'},
            {Key: 'LambdaFunction', Value: 'nshmp-haz-s3-lambda'}
          ]
        }
      ],
      UserData: (Buffer.from(userDataScript)).toString('base64')
    }).promise().then((response) => {
      console.log('response = ', response);

      const instance = response.Instances[0];

      let message = `
        Starting AWS EC2 instance ${instance.InstanceType}.
        ${dateFormat(new Date(instance.LaunchTime), NshmpLambdaUtil.timeFormat)}.


        ----------
        EC2 Instance Metadata:
          - EC2 instance id: ${instance.InstanceId}.
          - EC2 instance type: ${instance.InstanceType}.
          - Current state: ${instance.State.Name}.


        ----------
        Trigger config file (${config.triggerFileName}):
        \n${config.rawTriggerConfig}

      `;

      const subject = `AWS nshmp-haz (job id: ${config.jobId})`;

      return NshmpLambdaUtil.sendNotification(subject, message, config);
    }).then(() => {
      return this.s3.putBucketCors({
        Bucket: config.triggerConfig.nshmp.s3Bucket.split('/')[0],
        CORSConfiguration: {
          CORSRules: [
            {
              AllowedOrigins: ['*'],
              AllowedMethods: ['GET'],
              AllowedHeaders: ['*']
            }
          ]
        }
      }).promise();
    }).catch((error: Error) => {
      throw new LambdaError(error.message, config);
    });

  }

}
